# defines patterns which can be checked against for rule application
# SinglePattern checks for a match in either the unit or the name
# returns tuple (0,x) (1,x) (2,x)
# 0 means no match
# 1 means match in unit
# 2 means match in name
# x is index of the match in the sample
# DoublePattern checks for a match in both the unit and name

from .FormatCheck import listOfStrings, dictStrBool
from .Ingredient import Ingredient

class Pattern(object):
    def __init__(self):
        raise Exception("creating an instance of Pattern base class is forbidden")
    @staticmethod
    def _compare(rule,sample):
        # rule and sample are both lists of strings
        # want to know if sample matches rule
        # for a match, rule must be a subset of sample
        # returns the index of sample where rule begins
        # or returns None if no match
        for sample_index in range(len(sample)-len(rule)+1):
            matches = True
            for rule_index in range(len(rule)):
                if rule[rule_index] != sample[sample_index+rule_index]:
                    matches = False
                    break
            if matches:
                return sample_index
        return None
    @staticmethod
    def _check_props(rule,sample):
        # rule is a dict(str->bool)
        # sample is a set(str)
        # we want to make sure that all the
        # elements marked as True appear in the sample
        # and none of those marked False appear
        # returns True for a match, False otherwise
        for prop in rule:
            if rule[prop]:
                # it must exist in the set to match
                if prop not in sample:
                    return False
            else:
                # it cannot exist in the set to match
                if prop in sample:
                    return False
        return True

class SinglePattern(Pattern):
    def __init__(self,rule,props):
        if (rule is None) and (len(props) == 0):
            raise ValueError("must enter at least one criterion")
        if rule is not None:
            listOfStrings(rule)
        dictStrBool(props)
        self.rule = rule
        self.props = props
        self.specificity = len(props)
        if rule is not None:
            self.specificity += len(rule)
    def matches(self,ig):
        if not isinstance(ig,Ingredient):
            raise TypeError("Expected ig to be of type Ingredient")
        if len(self.props) == 0:
            res = self._compare(self.rule,ig.name)
            if res is not None:
                return (2,res)
            res = self._compare(self.rule,ig.unit)
            if res is not None:
                return (1,res)    
        else:
            # we need to check props
            # this means that we can
            # only match against the name
            if self.rule is None:
                res = -1
            else:
                res = self._compare(self.rule,ig.name)
            if res is not None:
                if self._check_props(self.props,ig.props):
                    return (2,res)
        return (0,None)

class DoublePattern(Pattern):
    def __init__(self,rule_unit,rule_name,props):
        # if rule_unit or rule_name are none, we interpret them as wildcards
        if (rule_unit is None) and (rule_name is None) and (len(props) == 0):
            raise ValueError("must enter at least one criterion")
        if rule_unit is not None:
            listOfStrings(rule_unit)
        if rule_name is not None:
            listOfStrings(rule_name)
        dictStrBool(props)
        self.rule_unit = rule_unit
        self.rule_name = rule_name
        self.props = props
        self.specificity = len(props)
        if rule_unit is not None:
            self.specificity += len(rule_unit)
        if rule_name is not None:
            self.specificity += len(rule_name)
    def matches(self,ig):
        if not isinstance(ig,Ingredient):
            raise TypeError("Expected ig to be of type Ingredient")
        if self.rule_unit is None:
            res_unit = -1
        else:
            res_unit = self._compare(self.rule_unit,ig.unit)
            if res_unit is None:
                return (0,None,None)
        if self.rule_name is None:
            res_name = -1
        else:
            res_name = self._compare(self.rule_name,ig.name)
            if res_name is None:
                return (0,None,None)
        if self._check_props(self.props,ig.props):
            return (1,res_unit,res_name)
        else:
            return (0,None,None)


# defines a class MyNumber
# this can hold floating point numbers
# and fractions, and convert between
# then easily

import math
from .SpudParser import SpudParser

class MyNumber(object):
    # has 2 member variables
    # bool is_fraction
    # ? value
    # if is_fraction is False, value will be a float
    # if is_fraction is True, value will be (int,int)
    def __init__(self,value):
        # value is either float
        # or tuple(int,int)
        if type(value) == float:
            if math.isnan(value) or math.isinf(value):
                raise ValueError("Expected value to be regular")
            if value < 0.0:
                raise ValueError("Expected value to be nonnegative")
            self.is_fraction = False
            self.value = value
        elif type(value) == tuple:
            if len(value) != 2:
                raise ValueError("Expected tuple to have length 2")
            if not all(map(lambda x: type(x) == int,value)):
                raise TypeError("Expected tuple to contain only int")
            if value[0] < 0:
                raise ValueError("Expected first int to be nonnegative")
            if value[1] < 1:
                raise ValueError("Expected second int to be greater than zero")
            self.is_fraction = True
            self.value = value
        else:
            raise TypeError("Expected float or tuple")
    def __str__(self):
        # returns a string representation of self.value
        # will use proper fractions over improper fractions
        if self.is_fraction:
            whole = self.value[0] // self.value[1]
            part = self.value[0] % self.value[1]
            if whole == 0:
                if part == 0:
                    # the value is zero
                    return "0"
                else:
                    # it's a short fraction
                    return str(part) + "/" + str(self.value[1])
            else:
                # it's an long fraction
                if part == 0:
                    # it's a whole number
                    return str(whole)
                else:
                    # it's a true long fraction
                    return str(whole) + " " + str(part) + "/" + str(self.value[1])
        else:
            return str(self.value)
    def as_float(self):
        # returns a version of this object
        # where is_fraction will be False
        if self.is_fraction:
            n = float(self.value[0]) / float(self.value[1])
            return MyNumber(n)
        else:
            return self
    def as_fraction(self,d):
        # returns a version of this object
        # where is_fraction will be True
        if self.is_fraction:
            return self
        # implements Farey sequence iteration
        if type(d) != int:
            raise TypeError("Expected d to be an int")
        if d < 1:
            raise ValueError("Expected d to be at least 1")
        whole = int(self.value)
        part = self.value - whole
        left = (0,1)
        right = (1,1)
        prev_best = left
        prev_best_distance = 2.0
        while True:
            now = (
                left[0] + right[0],
                left[1] + right[1]
            )
            if now[1] >= d:
                if now[1] == d:
                    return MyNumber(
                        (now[0]+whole*now[1],now[1])
                    )
                else:
                    return MyNumber(
                        (prev_best[0]+whole*prev_best[1],prev_best[1])
                    )
            # now[1] is too small
            f = float(now[0]) / float(now[1])
            if abs(f-part) < prev_best_distance:
                prev_best = now
                prev_best_distance = abs(f-part)
            if f == part:
                    return MyNumber(
                        (now[0]+whole*now[1],now[1])
                    )
            if f < part:
                # now is too small
                # it is the new lower bound
                left = now
            else:
                # now is too large
                # it is the new upper bound
                right = now
    def __add__(self,other):
        if not isinstance(other,MyNumber):
            raise TypeError("Expected other to be of type MyNumber")
        if self.is_fraction and other.is_fraction:
            # both are fractions
            # if they have the same denominator
            # this is a bit easier
            if self.value[1] == other.value[1]:
                # same denominators
                # can just add numerators
                n = self.value[0] + other.value[0]
                return MyNumber((n,self.value[1]))
            else:
                # different denominators
                # need to convert
                d = self.value[1] * other.value[1]
                n = self.value[0] * other.value[1] + other.value[0] * self.value[1]
                gcd = math.gcd(d,n)
                d = d // gcd
                n = n // gcd
                return MyNumber((n,d))
        else:
            # at least one of the inputs is not a fraction
            # we will return a floating point MyNumber
            n = self.as_float().value + other.as_float().value
            return MyNumber(n)
    def __mul__(self,other):
        if not isinstance(other,MyNumber):
            raise TypeError("Expected other to be of type MyNumber")
        if self.is_fraction and other.is_fraction:
            # both are fractions
            n = self.value[0] * other.value[0]
            d = self.value[1] * other.value[1]
            gcd = math.gcd(d,n)
            n = n // gcd
            d = d // gcd
            return MyNumber((n,d))
        else:
            # at least one is a floating-point number
            # return a floating-point MyNumber
            n = self.as_float().value * other.as_float().value
            return MyNumber(n)
    def multiplicative_inverse(self):
        # returns the multiplicative inverse, if it exists
        if self.is_fraction:
            return MyNumber( (self.value[1],self.value[0]) )
        else:
            return MyNumber( 1.0 / self.value )
    @staticmethod
    def _handle_fractionsimple(tree):
        # tree is SpudParser.FractionsimpleContext
        a = None
        for child in tree.children:
            if isinstance(child,SpudParser.NonzerostringContext):
                if a is None:
                    a = int(child.getText())
                else:
                    b = int(child.getText())
                    return (a,b)
        raise Exception("Internal Error")
    @staticmethod
    def _handle_fractioncomplex(tree):
        # tree is SpudParser.FractioncomplexContext
        a = None
        for child in tree.children:
            if isinstance(child,SpudParser.NonzerostringContext):
                a = int(child.getText())
            elif isinstance(child,SpudParser.FractionsimpleContext):
                b = MyNumber._handle_fractionsimple(child)
                return (b[0]+a*b[1],b[1])
        raise Exception("Internal Error")
    @staticmethod
    def from_tree(tree):
        # tree is SpudParser.NumberContext
        for child in tree.children:
            if isinstance(child,SpudParser.NonzerostringContext):
                return MyNumber((int(child.getText()),1))
            if isinstance(child,(SpudParser.DecimallargeContext,SpudParser.DecimalsmallContext)):
                return MyNumber(float(child.getText()))
            if isinstance(child,SpudParser.FractionsimpleContext):
                return MyNumber(MyNumber._handle_fractionsimple(child))
            if isinstance(child,SpudParser.FractioncomplexContext):
                return MyNumber(MyNumber._handle_fractioncomplex(child))
        raise Exception("Internal Error")

MyNumber.one = MyNumber((1,1))


# defines functions for checking format of inputs

def listOfStrings(x):
    # returns if x is a nonempty list
    # of nonempty strings
    # returns throws an exception otherwise
    if type(x) != list:
        raise TypeError("Expected list")
    if len(x) == 0:
        raise ValueError("Expected nonempty list")
    if not all(map(lambda y: (len(y) > 0) if (type(y) == str) else False , x)):
        raise TypeError("Expected elements of list to be nonempty strings")

def listOfTupleStringBools(x):
    if type(x) != list:
        raise TypeError("Expected list")
    for elem in x:
        if type(elem) != tuple:
            raise TypeError("Expected tuple")
        if len(elem) != 2:
            raise ValueError("Expected tuple of length 2")
        if type(elem[0]) != str:
            raise TypeError("Expected string")
        if type(elem[1]) != bool:
            raise TypeError("Expected bool")

def setOfStrings(x):
    # returns if x is a set of strings
    # throws and expection otherwise
    if type(x) != set:
        raise TypeError("Expected set")
    if not all(map(lambda y: (len(y) > 0) if (type(y) == str) else False , x)):
        raise TypeError("Expected elements of set to be nonempty strings")

def dictStrBool(x):
    # returns if x is dict(str->bool)
    # throws exception otherwise
    if type(x) != dict:
        raise TypeError("Expected dict")
    for key in x:
        if type(key) != str:
            raise ValueError("Expected dict with str keys")
        value = x[key]
        if type(value) != bool:
            raise TypeError("Expected dict with str keys and bool values")


# defines the SpudLoader class
# to load and parse source files

import os
import hashlib
from copy import copy
import antlr4
from antlr4.error.ErrorListener import ErrorListener
from .Pattern import SinglePattern, DoublePattern
from .Rule import Rule, HoldRule
from .RuleOutput import (
    RenamingRuleOutput, PrefixingRuleOutput, InsertingRuleOutput,
    SingleConvertingRuleOutput, DoubleConvertingRuleOutput,
    DecRuleOutputInstance, FracRuleOutputInstance, PropertiesRuleOutput)
from .SpudLexer import SpudLexer
from .SpudParser import SpudParser

from .MyNumber import MyNumber
from .Ingredient import Ingredient

class MyErrorListener(ErrorListener):
    # from https://stackoverflow.com/questions/32224980/python-2-7-antlr4-make-antlr-throw-exceptions-on-invalid-input
    def __init__(self,filename):
        super(MyErrorListener, self).__init__()
        self.filename = filename
    def syntaxError(self, recognizer, offendingSymbol, line, column, msg, e):
        raise Exception("Syntax Error in file: "+str(self.filename)+" on line: "+str(line)+" in column: "+str(column)+" with message: "+str(msg))
    def reportAmbiguity(self, recognizer, dfa, startIndex, stopIndex, exact, ambigAlts, configs):
        pass
    def reportAttemptingFullContext(self, recognizer, dfa, startIndex, stopIndex, conflictingAlts, configs):
        pass
    def reportContextSensitivity(self, recognizer, dfa, startIndex, stopIndex, prediction, configs):
        pass

class SpudLoader(object):
    def __init__(self):
        self.ingredients = []
        self.rules = []
        self.seen = set() # sha256 hashes of files
    @staticmethod
    def load_file_raw(path):
        if path[-5:] != ".spud":
            path = path + ".spud"
        if os.path.isfile(path):
            with open(path,"rb") as file:
                return file.read()
        else:
            raise Exception("Unable to locate file: "+path)
    def get_tree(self,abs_path):
        # abs_path is a string with the absolute file path to load
        # returns either the syntax tree for the source file
        # or None
        raw = self.load_file_raw(abs_path)
        file_hash = hashlib.sha256(raw).digest()
        if file_hash in self.seen:
            # we have seen this file before
            return None
        else:
            # we have not seen this file before
            # register it so that we do not
            # need to process it again
            self.seen.add(file_hash)
        # first convert the input to a string
        # and then to a stream
        raw_string = raw.decode("UTF-8")
        stream = antlr4.InputStream(raw_string)
        # stream contains the input
        lexer = SpudLexer(stream)
        token_stream = antlr4.CommonTokenStream(lexer)
        # token_stream contains a stream of tokens
        # from the input
        parser = SpudParser(token_stream)
        parser.addErrorListener( MyErrorListener(abs_path) )
        tree = parser.spud()
        # tree is the entire syntax tree for the input
        return tree
    @staticmethod
    def extract_imports(output,tree):
        if isinstance(tree,SpudParser.AtimportContext):
            # we found one
            output.append(tree.getText())
        elif isinstance(tree,(SpudParser.SpudContext,
                SpudParser.LineContext,
                SpudParser.LineatContext,
                SpudParser.AtsectionContext)):
            if tree.children is None:
                return
            for child in tree.children:
                SpudLoader.extract_imports(output,child)
    @staticmethod
    def convert_to_absolute_pathnames(base_dir,path):
        if os.path.isabs(path):
            return path
        else:
            output = os.path.join(base_dir,path)
            output = os.path.abspath(output)
            return output
    @staticmethod
    def resolve_import_path(base_dir,statement):
        # we know that @import exists in the statement
        index = statement.find("@import")
        # chop off the @import
        statement = statement[index+7:]
        # chop off all leading spaces
        while True:
            if len(statement) < 1:
                break
            if statement[0] == " " or statement[0] == "\t":
                statement = statement[1:]
            else:
                break
        # chop off all training spaces
        while True:
            if len(statement) < 1:
                break
            if statement[-1] == " " or statement[-1] == "\t":
                statement = statement[:-1]
            else:
                break
        return SpudLoader.convert_to_absolute_pathnames(base_dir,statement)
    def _handle_all_prop_elements(self,tree,context):
        # tree is SpudParser.Propspluselem
        # or SpudParser.Propselem
        # or SpudParser.Propswildelem
        for child in tree.children:
            # all terminal nodes
            text = child.getText()
            if child.getSymbol().type == SpudParser.Namesegment:
                return (text,sign)
            elif text == "$":
                return (text,sign)
            elif text == "+":
                sign = True
            elif text == "-":
                sign = False
    def _handle_all_props(self,tree,context):
        # tree is SpudParser.Propsplus
        # or SpudParser.Props
        # or SpudParser.Propswild
        if tree.children is None:
            return []
        props = []
        for child in tree.children:
            props.append(self._handle_all_prop_elements(child,context))
        return props
    @staticmethod
    def _props_conversion(props_in):
        # props_in is list(tuple(str,bool))
        # we will output the same information
        # in dict(str->bool)
        props_out = {}
        for element in props_in:
            props_out[element[0]] = element[1]
        return props_out
    def _handle_all_names(self,tree,context):
        # tree is SpudParser.Basicname
        # or SpudParser.Powname
        name = []
        if isinstance(tree,SpudParser.PownameContext):
            for child in tree.children:
                if isinstance(child,antlr4.tree.Tree.TerminalNodeImpl):
                    # this can only be !
                    name.append("!")
                else:
                    # this can only be SpudParser.Basicname
                    name += self._handle_all_names(child,context)
        else:
            # all children are terminal nodes
            for child in tree.children:
                if child.getSymbol().type == SpudParser.Namesegment:
                    name.append(child.getText())
        return name
    def _handle_name_and_props(self,tree,context):
        # tree is:
        # SpudParser.Basicnameplusprops
        # SpudParser.Pownamewithprops
        # SpudParser.Wildcardwithprops
        # SpudParser.Basicnamewithpropswild
        # SpudParser.Wildcardwithpropswild
        # SpudParser.Powwildwithprops
        # SpudParser.Basicwildwithprops
        if isinstance(tree,(SpudParser.PowwildwithpropsContext,SpudParser.BasicwildwithpropsContext)):
            return self._handle_name_and_props(tree.children[0],context)
        if isinstance(tree.children[0],antlr4.tree.Tree.TerminalNodeImpl):
            # can only be $
            name = None
        else:
            name = self._handle_all_names(tree.children[0],context)
        props = self._handle_all_props(tree.children[1],context)
        return name, props
    def _handle_lineingredient(self,tree,context):
        # tree is SpudParser.Lineingredient
        number = None
        unit = None
        for child in tree.children:
            if isinstance(child,SpudParser.NumberContext):
                number = MyNumber.from_tree(child)
            elif isinstance(child,SpudParser.BasicnameContext):
                # units
                unit = self._handle_all_names(child,context)
            elif isinstance(child,SpudParser.BasicnamepluspropsContext):
                # name and props
                name, props_raw = self._handle_name_and_props(child,context)
                props = set()
                for prop in props_raw:
                    props.add(prop[0])
        if number is None:
            # unit is also None
            number = MyNumber.one
            unit = ["mealsworth"]
        elif unit is None:
            # unit is None
            # but number is defined
            unit = ["count"]
        if "multiply" in context:
            number = number * context["multiply"]
        self.ingredients.append( Ingredient(number,unit,name,props) )
    def _handle_rule_side_simple(self,tree,context):
        # tree is SpudParser.RuleleftsimpleContext
        # or SpudParser.RulerightsimpleContext
        # or SpudParser.Rulerightsimplewild
        # returns list of (name,props) tuples
        output = []
        for child in tree.children:
            if isinstance(child,(
                    SpudParser.PownamewithpropsContext,
                    SpudParser.BasicnamewithpropswildContext,
                    SpudParser.BasicwildwithpropsContext)):
                output.append(self._handle_name_and_props(child,context))
        return output
    def _handle_line_renaming_prefixing(self,tree,context,output_type):
        # tree is SpudParser.LinerenamingContext
        # or SpudParser.LineprefixingContext
        # output_type is either RenamingRuleOutput or PrefixingRuleOutput
        for child in tree.children:
            if isinstance(child,SpudParser.RuleleftsimpleContext):
                left = self._handle_rule_side_simple(child,context)
            elif isinstance(child,SpudParser.RulerightsimplewildContext):
                right = self._handle_rule_side_simple(child,context)
        for i in range(len(right)):
            ro = output_type(right[i][0])
            if len(right[i][1]) != 0:
                ro = PropertiesRuleOutput(ro,right[i][1])
            right[i] = ro
        for i in range(len(left)):
            pattern = SinglePattern(left[i][0],self._props_conversion(left[i][1]))
            self.rules.append(Rule(pattern,right))
    def _handle_lineinserting(self,tree,context):
        # tree is SpudParser.LineinsertingContext
        for child in tree.children:
            if isinstance(child,SpudParser.RuleleftsimpleContext):
                left = self._handle_rule_side_simple(child,context)
            elif isinstance(child,SpudParser.RulerightsimpleContext):
                right = self._handle_rule_side_simple(child,context)
        for i in range(len(left)):
            pattern = SinglePattern(left[i][0],self._props_conversion(left[i][1]))
            pattern_size = len(left[i][0])
            outputs = []
            for j in range(len(right)):
                ro = InsertingRuleOutput(pattern_size,right[j][0])
                if len(right[j][1]) != 0:
                    ro = PropertiesRuleOutput(ro,right[j][1])
                outputs.append(ro)
            self.rules.append(Rule(pattern,outputs))
    def _handle_singleconvertingleftelem(self,tree,context):
        # tree is SpudParser.SingleconvertingleftelemContext
        for child in tree.children:
            if isinstance(child,SpudParser.NumberContext):
                number = MyNumber.from_tree(child)
            elif isinstance(child,SpudParser.PownamewithpropsContext):
                name, props = self._handle_name_and_props(child,context)
                return (number,name,props)
    def _handle_singleconvertingrightelem(self,tree,context):
        # tree is SpudParser.SingleconvertingrightelemContext
        for child in tree.children:
            if isinstance(child,SpudParser.NumberContext):
                number = MyNumber.from_tree(child)
            elif isinstance(child,antlr4.tree.Tree.TerminalNodeImpl):
                # can only be $
                number = None
            elif isinstance(child,SpudParser.BasicwildwithpropsContext):
                name, props = self._handle_name_and_props(child,context)
                return (number,name,props)
    def _handle_singleconvertingleft(self,tree,context):
        # tree is SpudParser.SingleconvertingleftContext
        patterns = []
        for child in tree.children:
            if isinstance(child,SpudParser.singleconvertingleftelem):
                patterns.append(self._handle_singleconvertingleftelem(child,context))
        return patterns
    def _handle_singleconvertingright(self,tree,context):
        # tree is SpudParser.SingleconvertingrightContext
        outputs = []
        for child in tree.children:
            if isinstance(child,SpudParser.singleconvertingrightelem):
                outputs.append(self._handle_singleconvertingrightelem(child,context))
        return outputs
    def _handle_linesingleconverting(self,tree,context):
        # tree is SpudParser.LinesingleconvertingContext
        for child in tree.children:
            if isinstance(child,SpudParser.SingleconvertingleftContext):
                left = self._handle_singleconvertingleft(child,context)
            elif isinstance(child,SpudParser.SingleconvertingrightContext):
                right = self._handle_singleconvertingright(child,context)
        for i in range(len(left)):
            # left[i][0] is the number, not None
            # left[i][1] is the name
            # left[i][2] is the props
            denominator = left[i][0].multiplicative_inverse()
            pattern = SinglePattern(left[i][1],self._props_conversion(left[i][2]))
            outputs = []
            for j in range(len(right)):
                # elements of right are (number,name,props)
                # number may be None
                if right[j][0] is None:
                    ratio = MyNumber.one
                else:
                    ratio = denominator * right[j][0]
                ro = SingleConvertingRuleOutput(ratio,right[j][1])
                if len(right[j][2]) != 0:
                    ro = PropertiesRuleOutput(ro,right[j][2])
                outputs.append(ro)
            self.rules.append(Rule(pattern,outputs))
    def _handle_doubleconvertingleftsub_1_2(self,tree,context):
        # tree is SpudParser.Doubleconvertingleftsub1Context
        # or SpudParser.Doubleconvertingleftsub2Context
        for child in tree.children:
            if isinstance(child,antlr4.tree.Tree.TerminalNodeImpl):
                if child.getText() == "$":
                    unit = None
                    continue
            if isinstance(child,SpudParser.PownameContext):
                unit = self._handle_all_names(child,context)
            elif isinstance(child,
                (SpudParser.PownamewithpropsContext,
                SpudParser.PowwildwithpropsContext)):
                    name, props = self._handle_name_and_props(child,context)
                    return unit, name, props
    def _handle_doubleconvertingleftelem(self,tree,context):
        # tree is SpudParser.DoubleconvertingleftelemContext
        for child in tree.children:
            if isinstance(child,SpudParser.NumberContext):
                number = MyNumber.from_tree(child)
            elif isinstance(child,
                (SpudParser.Doubleconvertingleftsub1Context,
                SpudParser.Doubleconvertingleftsub2Context)):
                    unit, name, props = self._handle_doubleconvertingleftsub_1_2(child,context)
                    return (number,unit,name,props)
    def _handle_doubleconvertingrightelem(self,tree,context):
        # tree is SpudParser.DoubleconvertingrightelemContext
        is_number_seen = False
        for child in tree.children:
            if isinstance(child,antlr4.tree.Tree.TerminalNodeImpl):
                # can only be $
                # but may be number or unit
                # depending on context
                if is_number_seen:
                    # it is the unit
                    unit = None
                else:
                    # it is the number
                    is_number_seen = True
                    number = None
            elif isinstance(child,SpudParser.NumberContext):
                is_number_seen = True
                number = MyNumber.from_tree(child)
            elif isinstance(child,SpudParser.BasicnameContext):
                unit = self._handle_all_names(child,context)
            elif isinstance(child,SpudParser.BasicwildwithpropsContext):
                name, props = self._handle_name_and_props(child,context)
                return (number,unit,name,props)
    def _handle_doubleconvertingleft(self,tree,context):
        # tree is SpudParser.DoubleconvertingleftContext
        patterns = []
        for child in tree.children:
            if isinstance(child,SpudParser.DoubleconvertingleftelemContext):
                patterns.append(self._handle_doubleconvertingleftelem(child,context))
        return patterns
    def _handle_doubleconvertingright(self,tree,context):
        # tree is SpudParser.DoubleconvertingrightContext
        outputs = []
        for child in tree.children:
            if isinstance(child,SpudParser.DoubleconvertingrightelemContext):
                outputs.append(self._handle_doubleconvertingrightelem(child,context))
        return outputs
    def _handle_linedoubleconverting(self,tree,context):
        # tree is SpudParser.LinedoubleconvertingContext
        for child in tree.children:
            if isinstance(child,SpudParser.DoubleconvertingleftContext):
                left = self._handle_doubleconvertingleft(child,context)
            elif isinstance(child,SpudParser.DoubleconvertingrightContext):
                right = self._handle_doubleconvertingright(child,context)
        for i in range(len(left)):
            # left[i][0] is the number, not None
            # left[i][1] is the unit, may be None
            # left[i][2] is the name, may be None
            # left[i][3] is the props
            denominator = left[i][0].multiplicative_inverse()
            pattern = DoublePattern(left[i][1],left[i][2],self._props_conversion(left[i][3]))
            outputs = []
            for j in range(len(right)):
                if right[j][0] is None:
                    ratio = None
                else:
                    ratio = denominator * right[j][0]
                ro = DoubleConvertingRuleOutput(ratio,right[j][1],right[j][2])
                if len(right[j][3]) != 0:
                    ro = PropertiesRuleOutput(ro,right[j][3])
                outputs.append(ro)
            self.rules.append(Rule(pattern,outputs))
    def _handle_linestandardrule(self,tree,context):
        # tree is SpudParser.LinestandardruleContext
        for child in tree.children:
            if isinstance(child,SpudParser.LinerenamingContext):
                self._handle_line_renaming_prefixing(child,context,RenamingRuleOutput)
            elif isinstance(child,SpudParser.LineprefixingContext):
                self._handle_line_renaming_prefixing(child,context,PrefixingRuleOutput)
            elif isinstance(child,SpudParser.LineinsertingContext):
                self._handle_lineinserting(child,context)
            elif isinstance(child,SpudParser.LinesingleconvertingContext):
                self._handle_linesingleconverting(child,context)
            elif isinstance(child,SpudParser.LinedoubleconvertingContext):
                self._handle_linedoubleconverting(child,context)
    def _handle_atsub1(self,tree,context):
        # tree is SpudPArser.Atsub1Context
        for child in tree.children:
            if isinstance(child,SpudParser.PownamewithpropsContext):
                name, props = self._handle_name_and_props(child,context)
                return DoublePattern(None,name,self._props_conversion(props))
    def _handle_atsub2(self,tree,context):
        # tree is SpudParser.Atsub2Context
        name = None
        props = {}
        for child in tree.children:
            if isinstance(child,SpudParser.PownameContext):
                unit = self._handle_all_names(child,context)
            elif isinstance(child,SpudParser.PowwildwithpropsContext):
                name, props = self._handle_name_and_props(child,context)
        return DoublePattern(unit,name,self._props_conversion(props))
    def _handle_atsub(self,tree,context):
        # tree is SpudParser.AtsubContext
        for child in tree.children:
            if isinstance(child,SpudParser.Atsub1Context):
                return self._handle_atsub1(child,context)
            if isinstance(child,SpudParser.Atsub2Context):
                return self._handle_atsub2(child,context)
    def _handle_athold(self,tree,context):
        # tree is SpudParser.AtholdContext
        for child in tree.children:
            if isinstance(child,SpudParser.AtsubContext):
                pattern = self._handle_atsub(child,context)
                self.rules.append(HoldRule(pattern,0))
                # holds have priority 0
    def _handle_atholdunit(self,tree,context):
        # tree is SpudParser.AtholdunitContext
        for child in tree.children:
            if isinstance(child,SpudParser.AtsubContext):
                pattern = self._handle_atsub(child,context)
                self.rules.append(HoldRule(pattern,4))
                # holdunits have priority 4
    def _handle_atdec(self,tree,context):
        # tree is SpudParser.AtdecContext
        for child in tree.children:
            if isinstance(child,SpudParser.AtsubContext):
                pattern = self._handle_atsub(child,context)
                self.rules.append(Rule(pattern,[DecRuleOutputInstance]))
    def _handle_atfrac(self,tree,context):
        # tree is SpudParser.AtfracContext
        for child in tree.children:
            if isinstance(child,SpudParser.AtsubContext):
                pattern = self._handle_atsub(child,context)
                self.rules.append(Rule(pattern,[FracRuleOutputInstance]))
    def _handle_atbegin(self,tree,context):
        # tree is SpudParser.AtbeginContext
        # return a new context with the correct value of multiply
        for child in tree.children:
            if isinstance(child,SpudParser.NumberContext):
                value = MyNumber.from_tree(child)
                new_context = copy(context)
                if "multiply" in new_context:
                    value = new_context["multiply"] * value
                new_context["multiply"] = value
                return new_context
        raise Exception("Internal Error")
    def _handle_atsection(self,tree,context):
        # tree is SpudParser.AtsectionContext
        for child in tree.children:
            # ignore atend
            if isinstance(child,SpudParser.AtbeginContext):
                context = self._handle_atbegin(child,context)
            elif isinstance(child,SpudParser.LineContext):
                self._handle_line(child,context)
    def _handle_lineat(self,tree,context):
        # tree is SpudParser.LineatContext
        for child in tree.children:
            if isinstance(child,SpudParser.AtholdContext):
                self._handle_athold(child,context)
            elif isinstance(child,SpudParser.AtholdunitContext):
                self._handle_atholdunit(child,context)
            elif isinstance(child,SpudParser.AtdecContext):
                self._handle_atdec(child,context)
            elif isinstance(child,SpudParser.AtfracContext):
                self._handle_atfrac(child,context)
            elif isinstance(child,SpudParser.AtsectionContext):
                self._handle_atsection(child,context)
    def _handle_line(self,tree,context):
        # tree is a SpudParser.LineContext
        if tree.children is None:
            return
        for child in tree.children:
            if isinstance(child,SpudParser.LineingredientContext):
                self._handle_lineingredient(child,context)
            elif isinstance(child,SpudParser.LinestandardruleContext):
                self._handle_linestandardrule(child,context)
            elif isinstance(child,SpudParser.LineatContext):
                self._handle_lineat(child,context)
    def _handle_spud(self,tree):
        # tree is a SpudParser.SpudContext
        if tree.children is None:
            return
        for child in tree.children:
            if isinstance(child,SpudParser.LineContext):
                self._handle_line(child,{})
    def recursive_load_parse(self,abs_path):
        # abs_path is a string with the absolute file path to load
        tree = self.get_tree(abs_path)
        if tree is None:
            return
        imports = []
        self.extract_imports(imports,tree)
        for statement in imports:
            self.recursive_load_parse(self.resolve_import_path(
                os.path.dirname(abs_path),
                statement
            ))
        self._handle_spud(tree)



# defines the ResultsTree class
# used for storing the output of the program
# and for saving the results as an html document

from .Ingredient import Ingredient

class ResultsTree(object):
    def __init__(self):
        self.terms = []
        self.subtrees = {}
    def add_ingredient(self,ig):
        if ig is None:
            return
        if not isinstance(ig,Ingredient):
            raise TypeError("Expected ingredient")
        put_in_terms = False
        if len(ig.name) == 1:
            if ig.name[0] == "==here==":
                put_in_terms = True
        if put_in_terms:
            # we will put it in the terms
            # try to add it to an exist term if possible
            for i in range(len(self.terms)):
                term = self.terms[i]
                if term.unit == ig.unit and term.props == ig.props:
                    # we have a match
                    self.terms[i] = Ingredient(
                        term.count + ig.count,
                        term.unit,
                        term.name,
                        term.props
                    )
                    return
            # no match in the terms
            # add a new term
            self.terms.append(ig)
        else:
            # we will put it in the subtrees
            if ig.name[0] in self.subtrees:
                # we can hand it off to the subtree for processing
                st = self.subtrees[ig.name[0]]
            else:
                # the subtree does not exist yet
                # we need to make it
                st = ResultsTree()
                self.subtrees[ig.name[0]] = st
            new_ig = ig.duplicate()
            if len(new_ig.name) == 1:
                new_ig.name = ["==here=="]
            else:
                new_ig.name = new_ig.name[1:]
            st.add_ingredient(new_ig)
    def as_html(self,node_name):
        # node_name is the string key
        # used to identify this object
        # in the parent tree's dictionary
        output = ["<li>",node_name,"<ul>"]
        for term in self.terms:
            output.append("<li>")
            output.append(str(term.count))
            for unit in term.unit:
                output.append(" ")
                output.append(unit)
            props = list(term.props)
            props.sort()
            for prop in props:
                output.append(" ")
                output.append(prop)
            output.append("</li>")
        subtree_names = list(self.subtrees)
        subtree_names.sort()
        for subtree_name in subtree_names:
            subtree = self.subtrees[subtree_name]
            output.append(subtree.as_html(subtree_name))
        output.append("</ul>")
        output.append("</li>")
        return "".join(output)
    def as_html_document(self):
        output = [
            "<!DOCTYPE html>","<html>","<head>",
            "<meta charset=\"utf-8\" />",
            "</head>","<body>","<ul>"
            ]
        output.append(self.as_html("groceries"))
        output.append("</ul>")
        output.append("</body>")
        output.append("</html>")
        return "".join(output)



# this file defines the Ingredient class
# it can represent an item, its quantities

from .MyNumber import MyNumber
from .FormatCheck import listOfStrings, setOfStrings
from copy import copy

class Ingredient(object):
    def __init__(self,count,unit,name,props):
        if not isinstance(count,MyNumber):
            raise TypeError("count must be of type MyNumber")
        listOfStrings(unit)
        listOfStrings(name)
        setOfStrings(props)
        self.count = count
        self.unit = unit
        self.name = name
        self.props = props
    def duplicate(self):
        return Ingredient(
            self.count,self.unit[:],
            self.name[:],copy(self.props)
            )
    def strip(self):
        # remove all terms starting with _
        unit = list(filter(lambda x: x[0] != "_", self.unit))
        name = list(filter(lambda x: x[0] != "_", self.name))
        if len(unit) == 0 or len(name) == 0:
            return None
        props = set(filter(lambda x: x[0] != "_", self.props))
        return Ingredient(self.count,unit,name,props)